from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404
from django.views.generic import CreateView, FormView, ListView

from . import forms, models


class GroupList(ListView):
    model = models.Group
    paginate_by = 5
    template_name = 'groups/group_list.html'


class GroupDetail(ListView):
    model = models.Discussion
    paginate_by = 5
    template_name = 'groups/group_detail.html'

    def get_queryset(self):
        return super(GroupDetail, self).get_queryset().filter(group=self.group)

    def get_context_data(self, *args, **kwargs):
        context = super(GroupDetail, self).get_context_data(*args, **kwargs)
        context['group'] = self.group
        return context

    def dispatch(self, request, *args, **kwargs):
        self.group = get_object_or_404(models.Group, pk=self.kwargs['pk'])
        return super(GroupDetail, self).dispatch(request, *args, **kwargs)


class DiscussionCreate(FormView):
    form_class = forms.DiscussionCreate
    template_name = 'groups/discussion_form.html'

    def get_group(self):
        return get_object_or_404(models.Group, pk=self.kwargs['pk'])

    def get_context_data(self, *args, **kwargs):
        context = super(DiscussionCreate, self).get_context_data(*args, **kwargs)
        context['group'] = self.get_group()
        return context

    def get_success_url(self):
        return reverse('discussion-thread', kwargs={'pk': self.pk})

    def form_valid(self, form):
        discussion = models.Discussion.objects.create(
            creator=self.request.user,
            group=self.get_group(),
            name=form.cleaned_data['name'],
        )
        models.Comment.objects.create(
            body=form.cleaned_data['comment'],
            discussion=discussion,
            user=self.request.user,
        )
        self.pk = discussion.pk
        return super(DiscussionCreate, self).form_valid(form)


class DiscussionThread(CreateView):
    """
    Comment on a Discussion
    """
    model = models.Comment
    form_class = forms.AddComment
    template_name = 'groups/discussion_thread.html'

    def get_queryset(self):
        """Display only the comments attached to a given discussion,
        newest at the bottom.
        """
        return self.get_discussion().comments.all()

    def get_discussion(self):
        return models.Discussion.objects.get(pk=self.kwargs['pk'])

    def get_context_data(self, *args, **kwargs):
        """
        Attach the discussion and its existing comments to the context, so
        they can be displayed.
        """
        context = super(DiscussionThread, self).get_context_data(*args, **kwargs)
        context['comments'] = self.get_queryset()
        context['discussion'] = self.get_discussion()
        return context

    def form_valid(self, form):
        form.instance.user = self.request.user
        form.instance.discussion = self.get_discussion()
        return super(DiscussionThread, self).form_valid(form)


class DiscussionSubscribe(FormView):
    form_class = forms.DiscussionSubscribeForm
    template_name = 'groups/subscribe_button.html'

    def get_discussion(self):
        return get_object_or_404(models.Discussion, pk=self.kwargs['pk'])

    def get_form_kwargs(self):
        """
        Add a kwarg denoting whether the subscribe button will subscribe or
        unsubscribe the user from the attached discussion.
        """
        user = self.request.user
        discussion = self.get_discussion()
        subscribe = user not in discussion.subscribers.all()

        kwargs = super(DiscussionSubscribe, self).get_form_kwargs()
        kwargs['subscribe'] = subscribe
        return kwargs

    def form_valid(self, form):
        user = self.request.user
        discussion = self.get_discussion()

        if form.cleaned_data['subscribe']:
            discussion.subscribers.add(user)
        else:
            discussion.subscribers.remove(user)

        return super(DiscussionSubscribe, self).form_valid(form)

    def get_success_url(self):
        return self.get_discussion().get_absolute_url()



from incuna_test_utils.testcases.urls import URLTestCase

from .. import views


class TestGroupUrls(URLTestCase):
    pk = 42

    def test_group_list(self):
        self.assert_url_matches_view(
            views.GroupList,
            '/groups/',
            'group-list',
        )

    def test_group_detail(self):
        self.assert_url_matches_view(
            views.GroupDetail,
            '/groups/{}/'.format(self.pk),
            'group-detail',
            url_kwargs={'pk': self.pk}
        )

    def test_discussion_create(self):
        self.assert_url_matches_view(
            views.DiscussionCreate,
            '/groups/{}/new-discussion/'.format(self.pk),
            'discussion-create',
            url_kwargs={'pk': self.pk}
        )

    def test_discussion_detail(self):
        self.assert_url_matches_view(
            views.DiscussionThread,
            '/groups/discussions/{}/'.format(self.pk),
            'discussion-thread',
            url_kwargs={'pk': self.pk}
        )

    def test_discussion_subscribe(self):
        self.assert_url_matches_view(
            views.DiscussionSubscribe,
            '/groups/discussions/{}/subscribe/'.format(self.pk),
            'discussion-subscribe',
            url_kwargs={'pk': self.pk}
        )



from django.test import TestCase
from incuna_test_utils.compat import (
    Python2AssertMixin,
    wipe_id_fields_on_django_lt_17
)

from . import factories
from .. import models


class TestDiscussionManager(Python2AssertMixin, TestCase):
    def test_for_group_pk(self):
        group = factories.GroupFactory.create()
        discussion = factories.DiscussionFactory.create(group=group)
        factories.DiscussionFactory.create()

        results = models.Discussion.objects.for_group_pk(group.pk)
        self.assertCountEqual([discussion], results)


class TestCommentManager(Python2AssertMixin, TestCase):
    def test_for_group_pk(self):
        discussion = factories.DiscussionFactory.create()
        comment = factories.CommentFactory.create(discussion=discussion)
        factories.CommentFactory.create()

        results = models.Comment.objects.for_discussion_pk(discussion.pk)
        self.assertCountEqual([comment], results)


class TestGroup(Python2AssertMixin, TestCase):
    def test_fields(self):
        fields = models.Group._meta.get_all_field_names()
        expected = {
            'id',
            'name',
            'is_private',
            'moderators',
            'watchers',
            'members_if_private',

            # From Discussion
            'discussions',
        }
        self.assertCountEqual(fields, expected)

    def test_get_absolute_url(self):
        group = factories.GroupFactory.create()
        expected = '/groups/{}/'.format(group.pk)
        self.assertEqual(group.get_absolute_url(), expected)


class TestDiscussion(Python2AssertMixin, TestCase):
    def test_fields(self):
        fields = models.Discussion._meta.get_all_field_names()
        expected = wipe_id_fields_on_django_lt_17([
            'id',
            'name',
            'group',
            'group_id',
            'creator',
            'creator_id',
            'date_created',
            'subscribers',

            # From Comment
            'comments',
        ])
        self.assertCountEqual(fields, expected)

    def test_get_absolute_url(self):
        discussion = factories.DiscussionFactory.create()
        expected = '/groups/discussions/{}/'.format(discussion.pk)
        self.assertEqual(discussion.get_absolute_url(), expected)

    def test_ordering(self):
        discussion_one, discussion_two = factories.DiscussionFactory.create_batch(2)

        discussions = models.Discussion.objects.all()
        expected = [discussion_two, discussion_one]
        self.assertSequenceEqual(discussions, expected)


class TestComment(Python2AssertMixin, TestCase):
    def test_fields(self):
        fields = models.Comment._meta.get_all_field_names()
        expected = wipe_id_fields_on_django_lt_17([
            'id',
            'body',
            'discussion',
            'discussion_id',
            'user',
            'user_id',
            'date_created',
        ])
        self.assertCountEqual(fields, expected)

    def test_get_absolute_url(self):
        comment = factories.CommentFactory.create()
        expected = '/groups/discussions/{}/'.format(comment.discussion.pk)
        self.assertEqual(comment.get_absolute_url(), expected)



import datetime

import pytz
from django.core.urlresolvers import reverse
from django.http import Http404
from incuna_test_utils.compat import Python2AssertMixin

from . import factories
from .utils import RequestTestCase
from .. import models, views


class TestGroupList(Python2AssertMixin, RequestTestCase):
    view_class = views.GroupList

    def test_get(self):
        groups = factories.GroupFactory.create_batch(2)

        request = self.create_request()
        view = self.view_class.as_view()

        response = view(request)
        self.assertEqual(response.status_code, 200)
        object_list = response.context_data['object_list']
        self.assertCountEqual(groups, object_list)


class TestGroupDetail(RequestTestCase):
    view_class = views.GroupDetail

    def test_get(self):
        group = factories.GroupFactory.create()

        request = self.create_request()
        view = self.view_class.as_view()

        response = view(request, pk=group.pk)
        self.assertEqual(response.status_code, 200)
        detail_object = response.context_data['group']
        self.assertEqual(group, detail_object)


class TestDiscussionThread(Python2AssertMixin, RequestTestCase):
    view_class = views.DiscussionThread

    def make_datetime(self, year, month, day):
        return datetime.datetime(year, month, day, tzinfo=pytz.utc)

    def test_get(self):
        discussion = factories.DiscussionFactory.create()
        comment = factories.CommentFactory.create(discussion=discussion)
        factories.CommentFactory.create()

        request = self.create_request()
        view = self.view_class.as_view()

        response = view(request, pk=discussion.pk)
        self.assertEqual(response.status_code, 200)
        self.assertCountEqual(response.context_data['comments'], [comment])
        self.assertEqual(response.context_data['discussion'], discussion)

    def test_sorting(self):
        discussion = factories.DiscussionFactory.create()
        newer_comment = factories.CommentFactory.create(
            discussion=discussion,
            date_created=self.make_datetime(2014, 1, 3)
        )
        older_comment = factories.CommentFactory.create(
            discussion=discussion,
            date_created=self.make_datetime(2014, 1, 1)
        )

        request = self.create_request()
        view = self.view_class.as_view()

        response = view(request, pk=discussion.pk)
        self.assertEqual(response.status_code, 200)
        comments = [older_comment, newer_comment]
        self.assertCountEqual(response.context_data['comments'], comments)
        self.assertEqual(response.context_data['discussion'], discussion)

    def test_post(self):
        discussion = factories.DiscussionFactory.create()
        user = self.user_factory.create()
        data = {
            'body': 'I am a comment!'
        }

        # Hit the view, passing in the necessaries.
        request = self.create_request('post', user=user, data=data)
        view = self.view_class.as_view()
        view(request, pk=discussion.pk)

        # Assert that one comment was created with the appropriate properties.
        created_comment = models.Comment.objects.get(discussion=discussion)
        self.assertEqual(created_comment.body, data['body'])
        self.assertEqual(created_comment.discussion, discussion)
        self.assertEqual(created_comment.user, user)


class TestDiscussionCreate(RequestTestCase):
    view_class = views.DiscussionCreate

    def test_get(self):
        group = factories.GroupFactory.create()

        request = self.create_request()
        view = self.view_class.as_view()

        response = view(request, pk=group.pk)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context_data['group'], group)

    def test_post(self):
        group = factories.GroupFactory.create()
        user = self.user_factory.create()
        data = {
            'creator': user,
            'group': group,
            'name': 'Super Sensible Discussion',
            'comment': 'Super sensible comment.',
        }

        self.assertFalse(models.Discussion.objects.exists())

        request = self.create_request('post', user=user, data=data)
        view = self.view_class.as_view()
        response = view(request, pk=group.pk)

        discussion = models.Discussion.objects.all()
        self.assertEqual(len(discussion), 1)

        self.assertEqual(response.status_code, 302)
        expected = reverse('discussion-thread', kwargs={'pk': discussion[0].pk})
        self.assertEqual(response['Location'], expected)


class TestDiscussionSubscribe(RequestTestCase):
    view_class = views.DiscussionSubscribe

    def test_get_discussion(self):
        discussion = factories.DiscussionFactory.create()
        view = self.view_class()

        # Assert that the discussion is found.
        view.kwargs = {'pk': discussion.pk}
        self.assertEqual(discussion, view.get_discussion())

    def test_get_discussion_404(self):
        view = self.view_class()

        # Assert we get a 404 when we miss.
        view.kwargs = {'pk': 42424242}
        with(self.assertRaises(Http404)):
            view.get_discussion()

    def test_form_subscribe(self):
        discussion = factories.DiscussionFactory.create()
        user = self.user_factory.create()

        # Hit the view, passing in the necessaries.
        request = self.create_request('post', user=user, data={'subscribe': True})
        view = self.view_class.as_view()
        view(request, pk=discussion.pk)

        # Assert that the user is now subscribed.
        self.assertIn(user, discussion.subscribers.all())
        self.assertIn(discussion, user.subscribed_discussions.all())

    def test_form_unsubscribe(self):
        discussion = factories.DiscussionFactory.create()
        user = self.user_factory.create()

        # Subscribe the user.
        discussion.subscribers.add(user)

        # Hit the view, passing in the necessaries.
        request = self.create_request('post', user=user, data={'subscribe': False})
        view = self.view_class.as_view()
        view(request, pk=discussion.pk)

        # Assert that the user is now unsubscribed.
        self.assertNotIn(user, discussion.subscribers.all())
        self.assertNotIn(discussion, user.subscribed_discussions.all())

    def test_get_form_kwargs(self):
        discussion = factories.DiscussionFactory.create()
        user = self.user_factory.create()

        # Feed the correct data into the view, so we can call get_form_kwargs.
        request = self.create_request('get', user=user)
        view = self.view_class()
        view.request = request
        view.kwargs = {'pk': discussion.pk}

        # With an unsubscribed user, hitting this view should return
        # subscribe=True.
        self.assertTrue(view.get_form_kwargs()['subscribe'])

        # Subscribe the user, then check again.  We should return
        # subscribe=False now.
        discussion.subscribers.add(user)
        self.assertFalse(view.get_form_kwargs()['subscribe'])



from crispy_forms.bootstrap import FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit
from django import forms
from django.core.urlresolvers import reverse_lazy

from . import models


class AddComment(forms.ModelForm):
    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.label_class = 'col-lg-2'
    helper.field_class = 'col-lg-8'
    helper.layout = Layout(
        'body',
        FormActions(
            Submit('submit', 'Post comment'),
        ),
    )

    class Meta:
        fields = ('body',)
        model = models.Comment


class DiscussionCreate(forms.Form):
    comment = forms.CharField(widget=forms.Textarea)
    name = forms.CharField(max_length=255)
    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.label_class = 'col-lg-2'
    helper.field_class = 'col-lg-8'
    helper.layout = Layout(
        'name',
        'comment',
        FormActions(
            Submit('submit', 'Create Discussion'),
        ),
    )


class DiscussionSubscribeForm(forms.Form):
    subscribe = forms.BooleanField(widget=forms.HiddenInput(), required=False)

    helper = FormHelper()
    helper.form_class = 'form-horizontal'
    helper.action = reverse_lazy('discussion-subscribe')

    class Meta:
        fields = ('subscribe',)

    def __init__(self, *args, **kwargs):
        """
        Accept a value for the 'subscribe' variable, and build the layout
        to reflect the action the form will take.
        """
        to_subscribe = kwargs.pop('subscribe', True)

        # setdefault is like 'get', but if it misses, puts the specified
        # default into kwargs and returns *that* object.
        initial_values = kwargs.setdefault('initial', {})
        initial_values['subscribe'] = to_subscribe

        super(DiscussionSubscribeForm, self).__init__(*args, **kwargs)

        button_text = 'Subscribe' if to_subscribe else 'Unsubscribe'
        self.helper.layout = Layout(
            FormActions(
                Submit('submit', button_text),
            ),
        )



from django.conf import settings
from django.core.urlresolvers import reverse
from django.db import models
from django.utils import timezone


class DiscussionManager(models.Manager):
    def for_group_pk(self, pk):
        return self.get_queryset().filter(group_id=pk)


class CommentManager(models.Manager):
    def for_discussion_pk(self, pk):
        return self.get_queryset().filter(discussion_id=pk)


class Group(models.Model):
    """
    A model for a discussion group, similar to a forum.
    Groups can be private, in which case users have to request to join before
    they can read and/or watch the group.  The 'members' field stores such
    users, and isn't used by non-private groups.
    """
    name = models.CharField(max_length=255)
    is_private = models.BooleanField(default=False)
    moderators = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        related_name='moderated_groups',
    )
    watchers = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        related_name='watched_groups',
    )
    members_if_private = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        related_name='private_groups_joined',
    )

    def get_absolute_url(self):
        return reverse('group-detail', kwargs={'pk': self.pk})


class Discussion(models.Model):
    """A model for a discussion thread in a group."""
    name = models.CharField(max_length=255)
    group = models.ForeignKey('groups.Group', related_name='discussions')
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='created_discussions',
    )
    date_created = models.DateTimeField(default=timezone.now)
    subscribers = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        related_name='subscribed_discussions',
    )

    objects = DiscussionManager()

    class Meta:
        ordering = ['-date_created']

    def get_absolute_url(self):
        return reverse('discussion-thread', kwargs={'pk': self.pk})

    def get_latest_comment(self):
        return self.comments.latest('date_created')

    def get_total_replies(self):
        return self.comments.count()


class Comment(models.Model):
    """A model for a comment in a discussion thread."""
    body = models.TextField()
    discussion = models.ForeignKey('groups.Discussion', related_name='comments')
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='comments')
    date_created = models.DateTimeField(default=timezone.now)

    objects = CommentManager()

    class Meta:
        ordering = ('date_created',)

    def get_absolute_url(self):
        return reverse('discussion-thread', kwargs={'pk': self.discussion.pk})



from incuna_test_utils.compat import Python2AssertMixin

from .utils import RequestTestCase
from .. import forms, models


class TestAddCommentForm(Python2AssertMixin, RequestTestCase):
    form = forms.AddComment
    model = models.Comment

    def test_form_fields(self):
        expected = ['body']
        fields = self.form.base_fields.keys()
        self.assertCountEqual(fields, expected)

    def test_form_valid(self):
        data = {'body': 'I am a comment'}

        form = self.form(data=data)
        self.assertTrue(form.is_valid(), msg=form.errors)

    def test_form_not_valid(self):
        form = self.form(data={})
        self.assertFalse(form.is_valid())
        self.assertIn('body', form.errors)


class TestDiscussionCreateForm(Python2AssertMixin, RequestTestCase):
    form = forms.DiscussionCreate

    def test_form_fields(self):
        expected = ['name', 'comment']
        fields = self.form.base_fields.keys()
        self.assertCountEqual(fields, expected)

    def test_form_valid(self):
        data = {
            'name': "What's in a name?",
            'comment': 'First!!!!!!',
        }

        form = self.form(data=data)
        self.assertTrue(form.is_valid(), msg=form.errors)

    def test_form_not_valid(self):
        form = self.form(data={})
        self.assertFalse(form.is_valid())
        self.assertIn('name', form.errors)


class TestDiscussionSubscribeForm(Python2AssertMixin, RequestTestCase):
    form = forms.DiscussionSubscribeForm

    def test_form_fields(self):
        expected = ['subscribe']
        fields = self.form.base_fields.keys()
        self.assertCountEqual(fields, expected)

    def test_form_valid(self):
        form = self.form(data={})
        self.assertTrue(form.is_valid(), msg=form.errors)

    def test_initial(self):
        form = self.form(subscribe=True)
        self.assertTrue(form.initial['subscribe'])
        form = self.form(subscribe=False)
        self.assertFalse(form.initial['subscribe'])
